use clap::Parser;
use cli::{check_prerequisites, get_full_path_dir};
use config::{read_config, ChartExtended, Config, ConfigImpl};
use helm::chart::Chart;
use output::{message_empty, message_error};
use std::{error::Error, process::exit};

use crate::mirror::mirror_from_mirror_obj;

mod config;
mod extensions;
mod mirror;
mod patches;
pub mod cli;
pub mod git;
pub mod helm;
pub mod include;
pub mod output;
pub mod template;
pub mod workdir;

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
    /// Name of the working dir
    #[arg(short, long)]
    workdir: Option<String>,
    /// Configure the chart section from cli
    /// If only one chart is needed
    #[arg(long)]
    chart: Option<String>,
    #[arg(long)]
    version_override: Option<String>,
    /// Path to the configuration file
    #[arg(short, long)]
    config: String,
    /// Dry run
    #[arg(short, long, default_value = "false")]
    dry_run: bool,
    #[arg(long, default_value = "false")]
    skip_prerequisites_check: bool,
    /// Init git patch. Use it if you want to create git patch for a chart
    /// It's going to pull a chart and init a git repo there, so you can
    /// apply changes and create a patch file
    /// It's not going to try mirroring changes, but will apply extensions
    /// and patches that are already defined
    #[arg(long)]
    init_git_patch: Option<Vec<String>>,
    #[arg(long, default_value = "helm")]
    helm_bin: String,
    #[arg(long, default_value = "git")]
    git_bin: String,
    #[arg(long, default_value = "yq")]
    yq_bin: String,
}

fn exec(args: Args) -> Result<(), Box<dyn Error>> {
    let prerequisites = vec![args.helm_bin, args.git_bin, args.yq_bin];
    check_prerequisites(prerequisites)?;
    let workdir_path = workdir::setup_workdir(args.workdir)?;
    let mut config: Config = read_config(args.config.clone())?;
    let config_full_path = get_full_path_dir(args.config.clone())?;
    if let Some(chart_arg) = args.chart {
        config.charts = config.charts.into_iter().filter(|chart| chart.chart.name == chart_arg).collect::<Vec<ChartExtended>>();
        if let Some(version_arg) = args.version_override {
            config.charts.iter_mut()
                .for_each(|chart| chart.chart.version = version_arg.clone());
        };
    };
    config.apply_includes(config_full_path.clone())?;
    config
        .charts
        .into_iter()
        .try_for_each(|mut chart| -> Result<(), Box<dyn Error>> {
            chart.populate_variables(config.variables.clone());
            let mut chart_mirrors = config.mirrors.clone();
            chart_mirrors.retain(|mirror| mirror.check_chart(&chart) == false);

            if chart_mirrors.is_empty() {
                output::message_info(&format!("mirrors are up-to-date for {}", chart.chart.name.clone()));
                return Ok(());
            }
            // First step is to pull the chart to the working dir
            let current_repo = chart.chart.find_repo(config.repositories.clone())?;
            let chart_path = current_repo.pull_chart(chart.chart.clone(), workdir_path.clone())?;
            chart.chart.version = current_repo.get_version(chart_path.clone())?;
            if let Some(extensions) = chart.extensions.clone() {
                extensions
                    .into_iter()
                    .try_for_each(|extension| -> Result<(), Box<dyn Error>> {
                        extension.apply(chart_path.clone(), config_full_path.clone())
                    })?
            }
            if let Some(patches) = chart.patches.clone() {
                patches
                    .into_iter()
                    .try_for_each(|patch| -> Result<(), Box<dyn Error>> {
                        patch.apply(chart_path.clone(), config.patches.clone())
                    })?
            }
            chart_mirrors
                .iter()
                .try_for_each(|mirror| -> Result<(), Box<dyn Error>> {
                    if chart.chart.mirrors.contains(&mirror.name) {
                        mirror_from_mirror_obj(mirror.clone())?.push(
                            workdir_path.clone(),
                            chart_path.clone(),
                            chart.clone(),
                            args.dry_run,
                        )?
                    };
                    Ok(())
                })?;
            message_empty(&format!("{}", chart.chart.name));
            Ok(())
        })?;
    Ok(())
}

fn main() {
    match exec(Args::parse()) {
        Ok(()) => message_empty("Thanks for using helmule"),
        Err(err) => {
            message_error(err);
            exit(1)
        }
    }
}
